/* Include all of the XPM files */

#ifdef HAVE_XPM

#ifndef XC_WIN32
#include <X11/xpm.h>
#endif

#include "lib/pixmaps/a.xpm"
#include "lib/pixmaps/b.xpm"
#include "lib/pixmaps/bd.xpm"
#include "lib/pixmaps/ccw.xpm"
#include "lib/pixmaps/co.xpm"
#include "lib/pixmaps/cp.xpm"
#include "lib/pixmaps/cw.xpm"
#include "lib/pixmaps/d2.xpm"
#include "lib/pixmaps/e.xpm"
#include "lib/pixmaps/fi.xpm"
#include "lib/pixmaps/fx.xpm"
#include "lib/pixmaps/pz.xpm"
#include "lib/pixmaps/li.xpm"
#include "lib/pixmaps/mk.xpm"
#include "lib/pixmaps/mv.xpm"
#include "lib/pixmaps/po2.xpm"
#include "lib/pixmaps/pu2.xpm"
#include "lib/pixmaps/i.xpm"
#include "lib/pixmaps/s.xpm"
#include "lib/pixmaps/t.xpm"
#include "lib/pixmaps/uj.xpm"
#include "lib/pixmaps/fy.xpm"
#include "lib/pixmaps/yp.xpm"
#include "lib/pixmaps/z4.xpm"
#include "lib/pixmaps/z5.xpm"
#include "lib/pixmaps/pl.xpm"
#include "lib/pixmaps/pm.xpm"
#include "lib/pixmaps/pn.xpm"
#include "lib/pixmaps/w.xpm"

#endif
