#ifndef RESOURCE_H
#define RESOURCE_H

//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by dialogs.rc
//
#define IDD_INPUTDIALOG                 101
#define IDD_HELPDLG			103
#define IDD_OUTPUTDLG                   104
#define IDC_EDIT1                       1000
#define IDC_INFOTEXT                    1001
#define IDC_HELPPIX			1003
#define IDC_PAGENUM                     1004
#define IDC_FILENAME                    1005
#define IDC_APPLY1                      1006
#define IDC_PAGELABEL                   1007
#define IDC_APPLY2                      1008
#define IDC_SCALE                       1009
#define IDC_APPLY3                      1010
#define IDC_WIDTH                       1011
#define IDC_APPLY4                      1012
#define IDC_HEIGHT                      1013
#define IDC_APPLY5                      1014
#define IDC_APPLY6                      1015
#define IDC_ORIENTATION                 1018
#define IDC_AUTOFIT                     1019
#define IDC_MODE                        1020
#define IDC_FPEDIT                      1021
#define IDC_PSINFO                      1022
#define IDC_STATIC1                     1023
#define IDC_STATIC2                     1024
#define IDC_STATIC3                     1025
#define IDC_STATIC4                     1026
#define IDC_STATIC5                     1027
#define IDC_STATIC6                     1028
#define IDC_STATIC7                     1029
#define IDC_FILEOPEN                    1030

#define IDR_XCIRCUIT                    2001

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        102
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         1001
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif

#define IDR_A	2001

#endif /* RESOURCE_H */
