#--------------------------------
# XCircuit Resource setup for TCL
#--------------------------------

option add *background #a07650 startupFile
option add *borderwidth 1 startupFile

option add *mainarea.corner.background brown4 startupFile
option add *message.background beige startupFile
option add *message1.background beige startupFile
option add *message2.background beige startupFile

option add *xcircuit.width 1200 startupFile
option add *xcircuit.height 930 startupFile
option add *xcircuit*activeBackground #a07650 startupFile
option add *xcircuit*foreground #a07650 startupFile
option add *xcircuit*activeForeground beige startupFile
option add *xcircuit*font -*-times-medium-r-normal--12-* startupFile

option add *xcircuit1.width 900 startupFile
option add *xcircuit1.height 700 startupFile
option add *xcircuit1*activeBackground #a07650 startupFile
option add *xcircuit1*foreground #a07650 startupFile
option add *xcircuit1*activeForeground beige startupFile
option add *xcircuit1*font -*-times-medium-r-normal--12-* startupFile

option add *xcircuit2.width 700 startupFile
option add *xcircuit2.height 500 startupFile
option add *xcircuit2*activeBackground #a07650 startupFile
option add *xcircuit2*foreground #a07650 startupFile
option add *xcircuit2*activeForeground beige startupFile
option add *xcircuit2*font -*-times-medium-r-normal--12-* startupFile

option add *menubar.filebutton*background beige startupFile
option add *menubar.editbutton*background beige startupFile
option add *menubar.textbutton*background beige startupFile
option add *menubar.optionsbutton*background beige startupFile
option add *menubar.windowbutton*background beige startupFile
option add *menubar.netlistbutton*background beige startupFile
option add *menubar*font -*-times-bold-r-normal--12-* startupFile

option add *librarymenu*activeBackground #805630 startupFile
option add *librarymenu*foreground #805630 startupFile
option add *librarymenu*activeForeground beige startupFile
option add *librarymenu*font -*-times-medium-r-normal--12-* startupFile
option add *librarymenu*background beige startupFile

option add *pagemenu*activeBackground #805630 startupFile
option add *pagemenu*foreground #805630 startupFile
option add *pagemenu*activeForeground beige startupFile
option add *pagemenu*font -*-times-medium-r-normal--12-* startupFile
option add *pagemenu*background beige startupFile

option add *filelist*activeBackground #805630 startupFile
option add *filelist*foreground #805630 startupFile
option add *filelist*background white startupFile
option add *filelist*activeForeground beige startupFile
option add *filelist*font -*-times-medium-r-normal--12-* startupFile
option add *filelist*txt*font -*-courier-medium-r-normal--12-* startupFile

option add *output*activeBackground #805630 startupFile
option add *output*foreground #805630 startupFile
option add *output*background white startupFile
option add *output*activeForeground beige startupFile
option add *output*font -*-times-medium-r-normal--12-* startupFile
option add *output*txt*font -*-courier-medium-r-normal--12-* startupFile

option add *output.textent.butp.psmenu*background beige startupFile
option add *output.textent.txtf.sizb.sizemenu*background beige startupFile
option add *output.textent.buto.orientmenu*background beige startupFile

option add *help*activeBackground #805630 startupFile
option add *help*foreground #805630 startupFile
option add *help*background white startupFile
option add *help*activeForeground beige startupFile
option add *help*font -*-times-medium-r-normal--12-* startupFile
option add *help*txt*font -*-times-courier-r-normal--12-* startupFile

option add *dialog*activeBackground #805630 startupFile
option add *dialog*foreground #805630 startupFile
option add *dialog*background white startupFile
option add *dialog*activeForeground beige startupFile
option add *dialog*font -*-times-medium-r-normal--12-* startupFile
option add *dialog*txt*font -*-courier-medium-r-normal--12-* startupFile

option add *makesymbol*activeBackground #805630 startupFile
option add *makesymbol*foreground #805630 startupFile
option add *makesymbol*background white startupFile
option add *makesymbol*activeForeground beige startupFile
option add *makesymbol*font -*-times-medium-r-normal--12-* startupFile
option add *makesymbol*txt*font -*-courier-medium-r-normal--12-* startupFile

option add *savetech*activeBackground #805630 startupFile
option add *savetech*foreground #805630 startupFile
option add *savetech*background white startupFile
option add *savetech*activeForeground beige startupFile
option add *savetech*font -*-times-medium-r-normal--12-* startupFile
option add *savetech*txt*font -*-courier-medium-r-normal--12-* startupFile

# End of Resource configuration file.
