/*************************************<+>*************************************
 *****************************************************************************
 **
 **   File:        Form.h
 **
 **   Project:     X Widgets
 **
 **   Description: Public include file for Form class widget
 **
 *****************************************************************************
 **   
 **   Copyright (c) 1988 by Hewlett-Packard Company
 **   Copyright (c) 1988 by the Massachusetts Institute of Technology
 **   
 **   Permission to use, copy, modify, and distribute this software 
 **   and its documentation for any purpose and without fee is hereby 
 **   granted, provided that the above copyright notice appear in all 
 **   copies and that both that copyright notice and this permission 
 **   notice appear in supporting documentation, and that the names of 
 **   Hewlett-Packard or  M.I.T.  not be used in advertising or publicity 
 **   pertaining to distribution of the software without specific, written 
 **   prior permission.
 **   
 *****************************************************************************
 *************************************<+>*************************************/

/*  Form Widget  */

extern WidgetClass XwformWidgetClass;

typedef struct _XwFormClassRec      * XwFormWidgetClass;
typedef struct _XwFormRec           * XwFormWidget;
typedef struct _XwFormConstraintRec * XwFormConstraints;

